@@ -1,10 +1,26 @@
-CGIparse <- function(string){
+CGIparse <- function(string, collapse = TRUE)
+{
+    boundary <- getMultiPartBoundary()  
+    if(length(boundary)) 
+       return(parseMultiPartFormData(boundary, string, splitLines = TRUE))
+      
     the.split.string <- lapply(strsplit(string, "&"), 
                              function(string)strsplit(string, "="))[[1]]
     arglist <- lapply(the.split.string, function(pair) pair[2])
     names(arglist) <- sapply(the.split.string, function(pair) pair[1])
-    lapply(arglist, hexDecode)}
-       
+    ans <- lapply(arglist, hexDecode)
+
+    if(collapse) {
+      w <- which(duplicated(names(ans)))
+      for(i in w) {
+         first <- match(names(ans)[i], names(ans))
+         ans[first] <- c(ans[first], ans[i])
+      }
+    }
+
+    ans
+}
+      
 hexDecode <- function(string){
     string <- gsub("\\+", " ", string)
     string <- gsub("%09", "\t", string)
@@ -17,13 +33,43 @@ hexDecode <- function(string){
               }
     pieces <- c(pieces[1], sapply(pieces[-1], dehex))
     paste(pieces, collapse = "")
-    }     
+}
 
-HTMLheader <- function(){
+
+HTMLheader <-
+function(title = character(0), css = character(0))
+{
     cat("<!doctype html public \"-//W3C/DTD HTML 4.0/EN\">\n")
     lf(2)
+
+    cat("<HTML>\n<HEAD>\n")
+    if(length(title))
+        cat("<title>", paste(title, collapse = " "), "</title>\n", sep = "")
+
+    if(length(css)) {
+        cssNames = names(css)
+        sapply(seq(along = css),
+               function(i) {
+                       cat('<LINK rel="StyleSheet" HREF="', css[i], '"',
+                           ifelse(length(cssNames) && cssNames[i] != "", paste(' MEDIA="', cssNames[i], '" ', sep = ""), ""),
+                           'TYPE="text/css">\n', sep = "")
+               })
     }
-         
+
+    
+    cat("</HEAD>\n\n<BODY>\n")
+}
+
+HTMLfooter =
+  #
+  # Output the closing </BODY></HTML>
+  #
+function()
+{
+  cat("</BODY></HTML>\n")
+}
+
+
 br <- function(n = 1){
     cat(paste(rep("<BR>", n), collapse = ""), "\n")
     }
@@ -47,37 +93,56 @@ tag <- function(tagname, ...)
 untag <- function(tagname){
     if (getOption("useUnquotedTagnames")) 
         tagname <- as.character(substitute(tagname))
-    cat("</", tagname, ">", sep = "")}
+    cat("</", tagname, ">", sep = "")
+}
   
 lf <- function(n = 1) cat(paste(rep("\n", n), sep = ""))
 
 comments <- function(text) cat("<!--", text, "-->")
 
 mailto <- function(text, address){
-    cat("<a href=\"mailto:", address, "\">", text, "</a>", sep="")}
+    cat("<a href=\"mailto:", address, "\">", text, "</a>", sep="")
+}
     
 linkto <- function(text, URL){
-    cat("<A href=\"", URL, "\">", text, "</A>", sep = "")}
-    
-webPNG <- function(file, ...){
-    if (!exists("graphDir")){
+    cat("<A href=\"", URL, "\">", text, "</A>", sep = "")
+}
+
+ 
+webPNG <- function(file, ..., graphDir){
+    if (missing(graphDir)) {
+      if(!exists("graphDir", envir = globalenv(), inherits = TRUE)) {
         cat("Error in webPNG(): graphDir not set\n\n")
         q("no") ##  abort if graphDir not specified
+      }
+      graphDir = get("graphDir", envir = globalenv(), inherits = TRUE)
     }
+
     if (!file.exists(graphDir)) {
         cat("Error in webPNG():", graphDir, "does not exist\n\n")
         q("no") ##  abort if specified graphDir does not exist
     }
     n <- nchar(graphDir)
     separator <- if (substr(graphDir, n, n) == "/") "" else "/"
-    bitmap(file = paste(graphDir, file, sep = separator), ...)
-    invisible(NULL)}
+
+    if(require(GDD, quietly = TRUE)) {
+       GDD(file = paste(graphDir, file, sep = separator), ...)
+    } else
+       bitmap(file = paste(graphDir, file, sep = separator), ...)
+    invisible(NULL)
+}
+
     
-img <- function (src, ...) 
+img <- function (src, ..., graphURLroot = "") 
 {
     result <- src
-    if (exists("graphURLroot")) 
-        result <- paste(graphURLroot, src, sep = "")
+    if(missing(graphURLroot)) {
+       if (exists("graphURLroot", envir = globalenv(), inherits = TRUE))
+         graphURLroot = get("graphURLroot", envir = globalenv(), inherits = TRUE)
+       else
+         graphURLroot = ""
+    }
+    result <- paste(graphURLroot, src, sep = "")
     result <- paste("<IMG SRC=\"", result,
                     "?nocache=", sample(9999, 1), "\"",
                     sep = "")
@@ -90,7 +155,7 @@ img <- function (src, ...)
         result <- paste(result, " ", dots, ">", sep = "")
     } else result <- paste(result, ">", sep = "")
     cat(result)
-    invisible("image")
+    invisible(result)
 }
 
 "ascii" <-
@@ -180,7 +245,7 @@ indentPrint <- function(object, indent = 4, ...){
     indent <- paste(rep(" ", indent), sep = "", collapse = "")
     cat(paste(indent, zz, sep = ""), sep = "\n")}
     
-.First.lib <- function(lib, pkg){
+.onLoad <- .First.lib <- function(lib, pkg){
     options(useUnquotedTagnames = TRUE)
     formData <<- Sys.getenv("FORM_DATA")
     if (formData == "") {  ## probably uncgi has been used
